"""SQL wrapper around SQLDatabase in langchain."""
from typing import Any, List, Optional
from overrides import override

from sqlalchemy import MetaData
from sqlalchemy.engine import Engine
from . import metadata
from llama_index.utilities.sql_wrapper import SQLDatabase
class CDMDatabase(SQLDatabase):


    def __init__(
        self,
        engine: Engine,
        schema: Optional[str] = None,
        ignore_tables: Optional[List[str]] = None,
        include_tables: Optional[List[str]] = None,
        sample_rows_in_table_info: int = 3,
        indexes_in_table_info: bool = False,
        custom_table_info: Optional[dict] = None,
        view_support: bool = False,
        max_string_length: int = 300,
    ):
        """Create engine from database URI."""
        self._engine = engine
        self._schema = schema
        if include_tables and ignore_tables:
            raise ValueError("Cannot specify both include_tables and ignore_tables")

        #! Inspect is not supported in SQL Alchemy 1.4. So getting info from metadata
        # self._inspector = inspect(self._engine)

        # including view support by adding the views as well as tables to the all
        # tables list if view_support is True
        self._all_tables = set(
           metadata.tables.keys()
        )

        self._include_tables = set(include_tables) if include_tables else set()
        if self._include_tables:
            missing_tables = self._include_tables - self._all_tables
            if missing_tables:
                raise ValueError(
                    f"include_tables {missing_tables} not found in database"
                )
        self._ignore_tables = set(ignore_tables) if ignore_tables else set()
        if self._ignore_tables:
            missing_tables = self._ignore_tables - self._all_tables
            if missing_tables:
                raise ValueError(
                    f"ignore_tables {missing_tables} not found in database"
                )
        usable_tables = self.get_usable_table_names()
        self._usable_tables = set(usable_tables) if usable_tables else self._all_tables

        if not isinstance(sample_rows_in_table_info, int):
            raise TypeError("sample_rows_in_table_info must be an integer")

        self._sample_rows_in_table_info = sample_rows_in_table_info
        self._indexes_in_table_info = indexes_in_table_info

        self._custom_table_info = custom_table_info
        if self._custom_table_info:
            if not isinstance(self._custom_table_info, dict):
                raise TypeError(
                    "table_info must be a dictionary with table names as keys and the "
                    "desired table info as values"
                )
            # only keep the tables that are also present in the database
            intersection = set(self._custom_table_info).intersection(self._all_tables)
            self._custom_table_info = {
                table: info
                for table, info in self._custom_table_info.items()
                if table in intersection
            }

        self._max_string_length = max_string_length

        self._metadata = metadata or MetaData()


        # including view support if view_support = true
        # self._metadata.reflect(
        #     views=view_support,
        #     bind=self._engine,
        #     only=list(self._usable_tables),
        #     schema=self._schema,
        # )


    @override
    def get_table_columns(self, table_name: str) -> List[Any]:
        """Get table columns."""
        return metadata.tables[table_name].columns.keys()

    @override
    def get_single_table_info(self, table_name: str) -> str:
        """Get table info for a single table."""
        # same logic as table_info, but with specific table names
        template = (
            "Table '{table_name}' has columns: {columns}, "
            "and foreign keys: {foreign_keys}."
        )
        columns = []
        # print(metadata.tables[table_name].foreign_keys)
        foreign_keys = []
        for column in metadata.tables[table_name].columns:
            columns.append(f"{column.name} ({column.type!s})")
            for foreign_key in column.foreign_keys:
                foreign_keys.append(
                    f"{column.name} -> "
                    f"{foreign_key.column.table.name}.{foreign_key.column.name}"
                )
        column_str = ", ".join(columns)
        foreign_key_str = ", ".join(foreign_keys)

        #     if column.get("comment"):
        #         columns.append(
        #             f"{column['name']} ({column['type']!s}): "
        #             f"'{column.get('comment')}'"
        #         )
        #     else:
        #         columns.append(f"{column['name']} ({column['type']!s})")

        # column_str = ", ".join(columns)
        # foreign_keys = []
        # for foreign_key in self._inspector.get_foreign_keys(table_name):
        #     foreign_keys.append(
        #         f"{foreign_key['constrained_columns']} -> "
        #         f"{foreign_key['referred_table']}.{foreign_key['referred_columns']}"
        #     )
        # foreign_key_str = ", ".join(foreign_keys)
        return template.format(
            table_name=table_name, columns=column_str, foreign_keys=foreign_key_str
        )




import asyncio
import pytest
import os
import datetime

@staticmethod
def test_create_cohort(pyomop_fixture, metadata_fixture, capsys):
    engine = pyomop_fixture.engine
    # create tables
    asyncio.run(pyomop_fixture.init_models(metadata_fixture))
    asyncio.run(create_llm_query(pyomop_fixture, engine))


async def create_llm_query(pyomop_fixture,engine):
    response = "I'm running in CI with no LLM"
    try:
        from src.pyomop import Cohort, CdmLLMQuery
        from llama_index.llms import Vertex
        from src.pyomop.llm_engine import CDMDatabase
        # Add a cohort
        async with pyomop_fixture.session() as session:
            async with session.begin():
                session.add(Cohort(cohort_definition_id=2, subject_id=100,
                    cohort_end_date=datetime.datetime.now(),
                    cohort_start_date=datetime.datetime.now()))
                await session.commit()

                # Use any LLM that llama_index supports
                llm = Vertex(
                    model="chat-bison",
                )
                sql_database = CDMDatabase(engine, include_tables=[
                    "cohort",
                ])
                query_engine = CdmLLMQuery(sql_database, llm=llm)

                response  = query_engine.query("Show each in table cohort with a subject id of 100?")
        await session.close()
        await engine.dispose()
    except ImportError:
        pass
    ## If we are running in CI, we don't have access to the LLM
    print(response)





from typing import Any
from llama_index.indices.struct_store.sql_query import (
    SQLTableRetrieverQueryEngine,
)
from llama_index.objects import (
    SQLTableNodeMapping,
    ObjectIndex,
    SQLTableSchema,
)
from llama_index import VectorStoreIndex
from llama_index import ServiceContext
from typing import Any, Optional
from llama_index.prompts import BasePromptTemplate
from llama_index.objects.base import ObjectRetriever
from langchain.embeddings import HuggingFaceEmbeddings
from .llm_engine import CDMDatabase

class CdmLLMQuery(SQLTableRetrieverQueryEngine):
    def __init__(
        self,
            sql_database: CDMDatabase,
            table_retriever: ObjectRetriever[SQLTableSchema] = None,
            text_to_sql_prompt: Optional[BasePromptTemplate] = None,
            context_query_kwargs: Optional[dict] = None,
            synthesize_response: bool = True,
            response_synthesis_prompt: Optional[BasePromptTemplate] = None,
            service_context: Optional[ServiceContext] = None,
            context_str_prefix: Optional[str] = None,
            sql_only: bool = False,
            llm: Optional[Any] = None, # FIXME: type
            similarity_top_k: int = 1,
            **kwargs: Any,
        ):
        self._sql_database = sql_database
        self._similarity_top_k = similarity_top_k
        embed_model = HuggingFaceEmbeddings(model_name="BAAI/bge-small-en-v1.5")

        if service_context is None:
            if llm is None:
                raise ValueError("Must provide either llm or service_context")
            service_context = ServiceContext.from_defaults(
                llm=llm,
                embed_model=embed_model,
            )
            self._llm = llm

        self._service_context = service_context

        self._table_node_mapping = SQLTableNodeMapping(sql_database)
        self._table_schema_objs = [
            (SQLTableSchema(table_name='care_site')),
            (SQLTableSchema(table_name='condition_occurrence')),
            (SQLTableSchema(table_name='cohort')),
            (SQLTableSchema(table_name='concept')),
            (SQLTableSchema(table_name='death')),
            (SQLTableSchema(table_name='device_exposure')),
            (SQLTableSchema(table_name='drug_exposure')),
            (SQLTableSchema(table_name='location')),
            (SQLTableSchema(table_name='measurement')),
            (SQLTableSchema(table_name='observation')),
            (SQLTableSchema(table_name='observation_period')),
            (SQLTableSchema(table_name='person')),
            (SQLTableSchema(table_name='procedure_occurrence')),
            (SQLTableSchema(table_name='provider')),
            (SQLTableSchema(table_name='visit_occurrence')),
            (SQLTableSchema(table_name='note'))
        ]  # add a SQLTableSchema for each table

        self._object_index = ObjectIndex.from_objects(
            self._table_schema_objs,
            self._table_node_mapping,
            VectorStoreIndex,
            service_context=self._service_context,
        )

        if table_retriever is None:
            table_retriever = self._object_index.as_retriever(similarity_top_k=similarity_top_k)

        self._table_retriever = table_retriever


        super().__init__(
            sql_database,
            table_retriever,
            text_to_sql_prompt,
            context_query_kwargs,
            synthesize_response,
            response_synthesis_prompt,
            service_context,
            context_str_prefix,
            sql_only,
            **kwargs,
        )




"""Example of using LLMs with PyOMOP
git clone https://github.com/dermatologist/pyomop.git@develop
cd pyomop
pip install -e .[llm]
"""

from pyomop import CdmEngineFactory, Cohort, metadata, CdmLLMQuery, CDMDatabase
import datetime
import asyncio
# Import any LLMs that llama_index supports and you have access to
# Require OpenAI API key to use OpenAI LLMs
from llama_index.llms import Vertex

async def main():
    # Create a sqllite database by default
    # You can also connect to an existing CDM database using the CdmEngineFactory
    cdm = CdmEngineFactory()
    # Postgres example below (db='mysql' also supported)
    # cdm = CdmEngineFactory(db='pgsql', host='', port=5432,
    #                       user='', pw='',
    #                       name='', schema='cdm6')

    engine = cdm.engine
    # Create Tables if required
    await cdm.init_models(metadata)

    async with cdm.session() as session:
        async with session.begin():

            # Adding  a cohort just for the example (not required if you have a OMOP CDM database)
            session.add(Cohort(cohort_definition_id=2, subject_id=100,
                cohort_end_date=datetime.datetime.now(),
                cohort_start_date=datetime.datetime.now()))
            await session.commit()

            # Use any LLM that llama_index supports
            llm = Vertex(
                model="chat-bison",
            )
            # Include tables that you want to query
            sql_database = CDMDatabase(engine, include_tables=[
                "cohort",
            ])
            query_engine = CdmLLMQuery(sql_database, llm=llm)
            # Try any complex query.
            response  = query_engine.query("Show each in table cohort with a subject id of 100?")
    print(response)
    """

| cohort_id | subject_id | cohort_name |
|---|---|---|
| 1 | 100 | Math |

    """
    # Close session
    await session.close()
    await engine.dispose()

# Run the main function
asyncio.run(main())